package com.xylink.config;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.google.common.base.Function;
import com.google.common.base.Optional;
import com.google.common.base.Predicate;
import com.xylink.log.*;
import com.xylink.model.AppToken;
import com.xylink.util.HttpUtil;
import com.xylink.util.Result;
import com.xylink.util.SignUtil;
import com.xylink.util.SslUtils;
import org.apache.commons.lang3.StringUtils;

import java.io.IOException;
import java.io.InputStream;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.concurrent.ConcurrentHashMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Created by wenya on 17/9/11.
 */
public class SDKConfigMgr {
    private static String sdkjarVersion = "3.10.2";
    private static final String CONFIG_FILE = "/com/xylink/config/config.json";
    private static final ObjectMapper objectMapper = new ObjectMapper();
    private static String DEFAULT_SERVER_HOST = "https://sdk.xylink.com";
    private static SDKConfig sdkConfig;
    private static XyLinkLog xyLinkLog;
    public static final String DEFAULT_ENTERPRISEID = "default";

    public static Map<String, SDKConfig> entConfigMap = new ConcurrentHashMap<String, SDKConfig>();

    private volatile static ConcurrentHashMap<String,Boolean> uaaSafeApiMap = new ConcurrentHashMap<String, Boolean>();

    /**
     * 初始化默认值
     */
    static {
        try {
            InputStream inputStream = SDKConfigMgr.class.getResourceAsStream(CONFIG_FILE);
            sdkConfig = objectMapper.readValue(inputStream, SDKConfig.class);
        } catch (Exception e) {
            sdkConfig = new SDKConfig();
        }

        if (sdkConfig == null) {
            sdkConfig = new SDKConfig();
            sdkConfig.setServerHost(DEFAULT_SERVER_HOST);
        }
        entConfigMap.put(DEFAULT_ENTERPRISEID, sdkConfig);
        setXyLinkLogType(XyLinkLogType.SystemOut);
        xyLinkLog.info("current sdk version: "+ sdkjarVersion);
    }

    //==================一般用于旧方式，并且是非通用方法的调用，此处作为保留不做删除==========
    //用的人应该很少了，但是仍需要保留

    /**
     * 默认使用 System 级别日志
     */
    public static SignatureVersion getSignatureVersion() {
        return sdkConfig.getSignatureVersion();
    }

    public static boolean enableSecurityHttpMethod() {
        return sdkConfig.isEnableSecurityHttp();
    }

    public static void setEnableSecurityHttp(boolean enableSecurityHttp) {
        sdkConfig.setEnableSecurityHttp(enableSecurityHttp);
    }

    public static String getServerHost(String enterpriseId) {
        SDKConfig config = SDKConfigMgr.entConfigMap.get(enterpriseId);
        SDKConfig realConfig = config == null ? SDKConfigMgr.entConfigMap.get(SDKConfigMgr.DEFAULT_ENTERPRISEID) : config;
        return realConfig.getServerHost();
    }

    public static String getServerHost(String enterpriseId, String clientId) {
        enterpriseId = enterpriseId.substring("enterpriseId=".length());

        //初始化SDKConfig使用企业的
        SDKConfig config = SDKConfigMgr.entConfigMap.get(enterpriseId);

        //如果有clientId，就直接使用clientId+enterpriseId进行查询，优先使用这个
        if (StringUtils.isNotEmpty(clientId)){
            config = SDKConfigMgr.entConfigMap.get(enterpriseId.concat(clientId));
        }

        //如果什么都没查到就使用默认的
        config = config == null ? SDKConfigMgr.entConfigMap.get(SDKConfigMgr.DEFAULT_ENTERPRISEID) : config;
        return config.getServerHost();
    }

    public static String getXyClientid() {
        return sdkConfig.getXyClientid();
    }

    public static String getAccessToken() {
        return sdkConfig.getAccessToken();
    }

    public static String getXySignSecret() {
        return sdkConfig.getXySignSecret();
    }

    public static void setServerHost(String host) {
        setServerHost(host, XyLinkLogType.SystemOut);
    }

    public static void setServerHost(String host, XyLinkLogType xyLinkLogType) {
        setXyLinkLogType(xyLinkLogType);
        if (host == null || host.trim().isEmpty()) {
            throw new IllegalArgumentException("host cannot be empty.");
        }
        if (!DEFAULT_SERVER_HOST.equals(host.toLowerCase().trim())) {
            try {
                SslUtils.ignoreSsl();
            } catch (Exception e) {
                throw new IllegalArgumentException("Failed to ignore ssl cert.", e);
            }
        }
        sdkConfig.setServerHost(host);
    }

    public static void addExtrasHeader(Map<String, String> extrasHeader) {
        if (extrasHeader == null || extrasHeader.isEmpty()) {
            return;
        }
        sdkConfig.setExtrasHeader(extrasHeader);
    }

    public static Map<String, String> getExtrasHeader() {
        return sdkConfig.getExtrasHeader();
    }

    public static void setSignType(SignUtil.SignType signType) {
        sdkConfig.setSignType(signType);
    }

    public static SignUtil.SignType getSignType() {
        return sdkConfig.getSignType();
    }
    //==================一般用于旧签名，签名1.0，并且是非通用方法的调用，此处作为保留不做删除==========


    //=================用户自定义配置==========================
    public static void initSDKConfig(String apiHost, String clientId, String clientSecret, String enterpriseId) {
        initSDKConfig(apiHost, clientId, clientSecret, enterpriseId, SignatureVersion.Signature_2_0);
    }

    public static void initSDKConfig(String apiHost, String clientId, String clientSecret, String enterpriseId, XyLinkLogType xyLinkLogType) {
        initSDKConfig(apiHost, clientId, clientSecret, enterpriseId, SignatureVersion.Signature_2_0, SignUtil.SignType.MD5, false, null, xyLinkLogType);
    }

    public static void initSDKConfigSign3(String apiHost, String clientId, String clientSecret, String enterpriseId, XyLinkLogType xyLinkLogType) {
        initSDKConfig(apiHost, clientId, clientSecret, enterpriseId, SignatureVersion.Signature_3_0, SignUtil.SignType.MD5, false, null, xyLinkLogType);
    }

    public static void initSDKConfig(String apiHost, String clientId, String clientSecret, String enterpriseId, SignatureVersion signatureVersion) {
        initSDKConfig(apiHost, clientId, clientSecret, enterpriseId, signatureVersion, SignUtil.SignType.MD5, false, null, XyLinkLogType.SystemOut);
    }

    public static void initSDKConfig(String apiHost, String clientId, String clientSecret, String enterpriseId, SignatureVersion signatureVersion,
                                     SignUtil.SignType signType,
                                     boolean enableSecurityHttp,
                                     Map<String, String> extrasHeader,
                                     XyLinkLogType xyLinkLogType) {
        setXyLinkLogType(xyLinkLogType);
        xyLinkLog.info(" ========== xylink sdk config init start ========== ");
        SDKConfig sdkConfig = new SDKConfig();
        sdkConfig.setSignatureVersion(signatureVersion);
        sdkConfig.setServerHost(apiHost);
        sdkConfig.setXyClientid(clientId);
        sdkConfig.setXyClientSecret(clientSecret);
        sdkConfig.setEnterpriseId(enterpriseId);
        sdkConfig.setSignType(signType);
        sdkConfig.setEnableSecurityHttp(enableSecurityHttp);
        sdkConfig.setExtrasHeader(extrasHeader);
        sdkConfig.check();
        initAccesstoken(sdkConfig, new Predicate<SDKConfig>() {
            @Override
            public boolean apply(SDKConfig input) {return false;}
        }, false);
        //存储数据
        entConfigMap.put(sdkConfig.getEnterpriseId(), sdkConfig);
        entConfigMap.put(sdkConfig.getEnterpriseId().concat(clientId), sdkConfig);
        xyLinkLog.info(" ========== xylink sdk config init end ========== ");
    }

    public static void initSDKConfig(List<EnterpriseSdkConfig> multiEnterpriseConfig, XyLinkLogType xyLinkLogType) {
        setXyLinkLogType(xyLinkLogType);
        xyLinkLog.info(" ========== xylink sdk config init start ========== ");
        for (EnterpriseSdkConfig enterpriseSdkConfig : multiEnterpriseConfig) {
            SDKConfig record = new SDKConfig(enterpriseSdkConfig);
            record.check();
            //只有大于1.0的版本才进行初始化
            initAccesstoken(record, new Predicate<SDKConfig>() {
                @Override
                public boolean apply(SDKConfig input) {return false;}
            },false);
            //存储数据
            entConfigMap.put(record.getEnterpriseId(), record);
            entConfigMap.put(enterpriseSdkConfig.getEnterpriseId().concat(enterpriseSdkConfig.getClientId()), record);
        }
        xyLinkLog.info(" ========== xylink sdk config init end ========== ");
    }

    public static SDKConfig initAccesstoken(SDKConfig sdkconfig, Predicate<SDKConfig> returnPredicate, boolean trylock) {
        xyLinkLog.info("start initAccesstoken");
        if (!trylock){
            //需要阻塞执行，多个线程阻塞等待
            try {
                sdkconfig.getReentrantLock().lock();
                xyLinkLog.info("sync initAccesstoken");
                getSdkConfig(sdkconfig, returnPredicate);
            } finally {
                if (sdkconfig.getReentrantLock().isLocked()){
                    sdkconfig.getReentrantLock().unlock();
                }
            }
            xyLinkLog.info("end initAccesstoken");
            return sdkconfig;
        }

        //如果此动作可以任意一个线程操作，那就使用trylock
        if (sdkconfig.getReentrantLock().tryLock()){
            try {
                xyLinkLog.info("async initAccesstoken");
                getSdkConfig(sdkconfig, returnPredicate);
            } finally {
                if (sdkconfig.getReentrantLock().isLocked()){
                    sdkconfig.getReentrantLock().unlock();
                }
            }
        }else {
            xyLinkLog.info("async initAccesstoken already locked, leave directly");
        }
        xyLinkLog.info("end initAccesstoken");
        return sdkconfig;
    }

    private static void getSdkConfig(SDKConfig sdkconfig, Predicate<SDKConfig> returnPredicate) {
        try {
            //doubleCheck
            if (returnPredicate.apply(sdkconfig)) {
                xyLinkLog.info("get lock, but initAccesstoken already execute by another thread, return directly");
                //如果达到先决条件，后面执行的将是重复的工作，可以直接返回
                return;
            }
            initOrFreshSdkConfig(sdkconfig);
            //获取时间
            Long timeLog = getTimeLog(sdkconfig);
            sdkconfig.setTimeLag(timeLog);
            //设置更新时间
            sdkconfig.setUpdateTime(System.currentTimeMillis());
            //存储数据
            entConfigMap.put(sdkconfig.getEnterpriseId(), sdkconfig);
            entConfigMap.put(sdkconfig.getEnterpriseId().concat(sdkconfig.getXyClientid()), sdkconfig);
        } catch (Exception e) {
            xyLinkLog.error("initAccesstoken error", e);
            throw new RuntimeException(e);
        }
    }

    private static void initOrFreshSdkConfig(SDKConfig sdkconfig) {
        //由于上游已经添加了锁，所以此处不用担心多线程的问题
        Boolean uaaSafeApi = uaaSafeApiMap.get(sdkconfig.getServerHost());
        //先使用新的接口进行访问，如果报错就使用旧的接口
        Result result = requestForResult(sdkconfig, uaaSafeApi);
        if (!result.isSuccess()) {
            xyLinkLog.error("uaaSafeApi: " + uaaSafeApiMap.get(sdkconfig.getServerHost()) + " ,initOrFreshSdkConfig fail by result " + result);
            //给个机会，重新尝试新接口
            uaaSafeApiMap.remove(sdkconfig.getServerHost());
            return;
        }

        xyLinkLog.info("uaaSafeApi: " + uaaSafeApiMap.get(sdkconfig.getServerHost()) +" , clientid: " + sdkconfig.getXyClientid() + " oauth appToken " + result);

        Map appTokenMap = (Map) result.getData();
        if (appTokenMap == null || appTokenMap.isEmpty() || !appTokenMap.get("message").equals("success") || appTokenMap.get("data") == null) {
            xyLinkLog.error("initOrFreshSdkConfig fail by appTokenMap " + appTokenMap);
            return;
        }
        sdkconfig.setAccessToken((String) ((Map) appTokenMap.get("data")).get("access_token"));
        sdkconfig.setRefreshToken((String) ((Map) appTokenMap.get("data")).get("refresh_token"));
        sdkconfig.setXySignSecret((String) ((Map) appTokenMap.get("data")).get("signSecret"));
        //获取过期时间
        Integer accessTokenExpireIn = Optional.fromNullable((Map) appTokenMap.get("data")).transform(new Function<Map, Integer>() {
            @Override
            public Integer apply(Map input) {
                Object o = input.get("expires_in");
                return o == null ? 0 : (Integer) o;
            }
        }).or(0);
        //获取时间戳
        Long timestamp = Optional.fromNullable((Map) appTokenMap.get("data")).transform(new Function<Map, Long>() {
            @Override
            public Long apply(Map input) {
                Object o = input.get("timestamp");
                return o == null ? System.currentTimeMillis() : (Integer) o;
            }
        }).or(System.currentTimeMillis());
        //计算过期时间
        sdkconfig.setAccessTokenExpireIn(accessTokenExpireIn * 1000 + timestamp);
        xyLinkLog.info("initOrFreshSdkConfig success " + sdkconfig);
    }

    private static Result requestForResult(SDKConfig sdkconfig, Boolean uaaSafeApi) {
        AppToken appToken = new AppToken();
        appToken.setClient_id(sdkconfig.getXyClientid());
        appToken.setEnterpriseId(sdkconfig.getEnterpriseId());
        try {
            if (Objects.isNull(uaaSafeApi)){
                //如果为空代表第一次使用所以要尝试使用哪个接口
                xyLinkLog.info("guess which uaa api can be use");
                try {
                    // 尝试新接口
                    Result result = getFromSafeApi(sdkconfig, appToken);
                    uaaSafeApiMap.put(sdkconfig.getServerHost(), true);
                    uaaSafeApi = true;
                    xyLinkLog.info("use uaa safe api");
                    return result;
                } catch (Exception e) {
                    //如果报错就使用旧接口，并且记录状态
                    //尝试旧接口
                    Result result = getFromUnSafeApi(sdkconfig, appToken);
                    xyLinkLog.info("use uaa unsafe api");
                    uaaSafeApiMap.put(sdkconfig.getServerHost(), false);
                    uaaSafeApi = false;
                    return  result;
                }
            }

            if (uaaSafeApi){
                return getFromSafeApi(sdkconfig, appToken);
            }else {
                return getFromUnSafeApi(sdkconfig, appToken);
            }

        } catch (Exception e) {
            xyLinkLog.error("uaaSafeApi: " + uaaSafeApiMap.get(sdkconfig.getServerHost()) + " ,initOrFreshSdkConfig fail", e);
            //给个机会，重新尝试新接口
            uaaSafeApiMap.remove(sdkconfig.getServerHost());
            throw new RuntimeException(e);
        }
    }

    private static Result getFromUnSafeApi(SDKConfig sdkconfig, AppToken appToken) throws IOException {
        appToken.setClient_secret(sdkconfig.getXyClientSecret());
        String jsonEntity = new ObjectMapper().writeValueAsString(appToken);
        return HttpUtil.executeHttp(sdkconfig.getServerHost() + "/admin/login/oauth/app_token", "POST", jsonEntity, Map.class, sdkconfig);
    }

    private static Result getFromSafeApi(SDKConfig sdkconfig, AppToken appToken) throws IOException {
        String jsonEntity = new ObjectMapper().writeValueAsString(appToken);
        Result result = HttpUtil.executeHttp(sdkconfig.getServerHost() + "/api/rest/gw/external/app_token/v1", "POST", jsonEntity, Map.class, sdkconfig, false);
        if (!result.isSuccess()) {
            //如果报错就使用旧接口，并且记录状态
            xyLinkLog.info("initOrFreshSdkConfig fail by /api/rest/gw/external/app_token/v1; result " + result);
            throw new IllegalStateException("initOrFreshSdkConfig fail by result " + result);
        }
        return result;
    }

    private static Long getTimeLog(SDKConfig sdkconfig) {
        try {
            //如果不是时间安全的版本就直接返回
            if (!sdkconfig.getSignatureVersion().isTimeSafeVersion()){
                return 0L;
            }
            Result sysinfo = HttpUtil.executeHttp(sdkconfig.getServerHost() + "/api/rest/uaa/external/sysinfo", "GET", "", Map.class, sdkconfig);
            if (sysinfo == null || !sysinfo.isSuccess()) {
                xyLinkLog.error("uaa sysinfo fail " + sysinfo);
                return 0L;
            }
            xyLinkLog.info("get uaa sysinfo " + sysinfo);
            Map sysinfoMap = (Map) sysinfo.getData();
            if (sysinfoMap == null || sysinfoMap.isEmpty() || sysinfoMap.get("data") == null || !sysinfoMap.get("message").equals("success")) {
                xyLinkLog.error("uaa sysinfoMap fail " + sysinfoMap);
                return 0L;
            }
            Long serverTimestamp = (Long) ((Map) sysinfoMap.get("data")).get("serverTimestamp");
            if (serverTimestamp == null) {
                xyLinkLog.error("uaa serverTimestamp is null ");
                return 0L;
            }
            Long local_init_time = System.currentTimeMillis();
            Long timeLag = serverTimestamp - System.currentTimeMillis();
            xyLinkLog.info(serverTimestamp + " - " + local_init_time + " = " + timeLag);
            return timeLag;
        } catch (Exception e) {
            xyLinkLog.error("get uaa sysinfo error ", e);
            return 0L;
        }
    }

    public static XyLinkLog getXyLinkLog() {
        return xyLinkLog;
    }

    /**
     * 默认使用 System 级别日志
     */
    public static void setXyLinkLogType(XyLinkLogType xyLinkLogType) {
        if (xyLinkLogType == null){
            xyLinkLog = SystemLogImpl.build();
            return;
        }

        switch (xyLinkLogType) {
            case SystemOut:
                xyLinkLog = SystemLogImpl.build();
                break;
            case Log4j:
                xyLinkLog = Log4jImpl.build();
                break;
            case Log4j2:
                xyLinkLog = Log4j2Impl.build();
                break;
            case LogBack:
                xyLinkLog = LogBackImpl.build();
                break;
            default:
                xyLinkLog = SystemLogImpl.build();
                break;
        }
    }

    private static final Pattern enterpriseIdPattern = Pattern.compile("(enterpriseId=\\w+)");

    /**
     * 依赖url查找config，优先使用clientid+enterpriseid的，然后是enterpriseid的，然后是默认的
     *
     * @param reqPath
     * @return
     */
    public static SDKConfig getSdkConfig(String reqPath, String clientId) {
        Matcher matcher = enterpriseIdPattern.matcher(reqPath);
        String enterpriseId = "";
        while (matcher.find()) {
            enterpriseId = matcher.group();
        }
        enterpriseId = enterpriseId.substring("enterpriseId=".length());

        SDKConfig config;
        //如果有clientId，就直接使用clientId+enterpriseId进行查询，优先使用这个
        if (StringUtils.isEmpty(clientId)){
            config = SDKConfigMgr.entConfigMap.get(enterpriseId);
        }else {
            config = SDKConfigMgr.entConfigMap.get(enterpriseId.concat(clientId));
        }

        //如果什么都没查到就使用默认的
        return config == null ? SDKConfigMgr.entConfigMap.get(SDKConfigMgr.DEFAULT_ENTERPRISEID) : config;
    }
}
