package com.xylink.sdk;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.xylink.config.SDKConfig;
import com.xylink.config.SDKConfigMgr;
import com.xylink.model.HttpMethod;
import com.xylink.util.HttpUtil;
import com.xylink.util.Result;
import com.xylink.util.SignatureSample;
import org.apache.commons.lang3.StringUtils;

import java.io.IOException;
import java.util.Map;
import java.util.TreeMap;

import static com.xylink.config.SignatureVersion.Signature_2_0_1;

/**
 * @Classname SdkMethod
 * @Date 2022/4/6 10:39 上午
 * @Created by anan
 * @Description sdkjar 通用方法，用于用户调用SDKserver的接口，用户需要传入url，body以及选择正确的请求方法即可
 * url的形式如下：/api/rest/xxx?xxx=xxxxx
 * 需要事先进行初始化
 *
 * #####################
 * 需要使用签名算法1.0   #
 * ####################
 */
public class SdkMethodV1 extends SdkMethodBase{

    /**
     * 使用签名方法1.0
     * @param url   /api/rest/xxx?xxx=xxxxx,不包含&signature=XXX
     * @param token 企业token
     * @return 返回文档描述中的对象
     * @throws IOException 抛出异常
     */
    public static <T> Result<T> sdkGet(String url, String token, Class<T> responseType) throws IOException {
        url = appendSignAndHost(url, token, StringUtils.EMPTY, HttpMethod.GET.name());
        return HttpUtil.getResponse(url, HttpMethod.GET.name(), StringUtils.EMPTY, responseType);
    }

    /**
     * @param url   /api/rest/xxx?xxx=xxxxx,不包含&signature=XXX
     * @param token 企业token
     * @param body  文档描述中的body内容
     * @return 返回文档描述中的对象
     * @throws IOException 抛出异常
     */
    public static <T> Result<T> sdkPost(String url, String token, Object body, Class<T> responseType) throws IOException {
        //body序列化
        String jsonEntity = new ObjectMapper().writeValueAsString(body);
        url = appendSignAndHost(url, token, jsonEntity, HttpMethod.POST.name());
        return HttpUtil.getResponse(url, HttpMethod.POST.name(), jsonEntity, responseType);
    }

    /**
     * @param url   /api/rest/xxx?xxx=xxxxx,不包含&signature=XXX
     * @param token 企业token
     * @param body  文档描述中的body内容
     * @return 返回文档描述中的对象
     * @throws IOException 抛出异常
     */
    public static <T> Result<T> sdkPut(String url, String token, Object body, Class<T> responseType) throws IOException {
        //body序列化
        String jsonEntity = new ObjectMapper().writeValueAsString(body);
        url = appendSignAndHost(url, token, jsonEntity, HttpMethod.PUT.name());
        return HttpUtil.getResponse(url, HttpMethod.PUT.name(), jsonEntity, responseType);
    }

    /**
     * @param url   /api/rest/xxx?xxx=xxxxx,不包含&signature=XXX
     * @param token 企业token
     * @return 返回文档描述中的对象
     * @throws IOException 抛出异常
     */
    public static <T> Result<T> sdkDelete(String url, String token, Class<T> responseType) throws IOException {
        url = appendSignAndHost(url, token, StringUtils.EMPTY, HttpMethod.DELETE.name());
        return HttpUtil.getResponse(url, HttpMethod.DELETE.name(), StringUtils.EMPTY, responseType);
    }

    /**
     * 携带body的delete方法
     * @param url
     * @param token
     * @param body
     * @param responseType
     * @return
     * @param <T>
     * @throws IOException
     */
    public static <T> Result<T> sdkDelete(String url, String token, Object body, Class<T> responseType) throws IOException {
        String jsonEntity = new ObjectMapper().writeValueAsString(body);
        url = appendSignAndHost(url, token, jsonEntity, HttpMethod.DELETE.name());
        return HttpUtil.getResponse(url, HttpMethod.DELETE.name(), jsonEntity, responseType);
    }
}
